package ims.coref.training;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ims.coref.data.Chain;
import ims.coref.data.Document;
import ims.coref.data.PairInstance;
import ims.coref.data.Span;
import ims.coref.data.SpanListStruct;
import ims.coref.markables.IMarkableExtractor;

public abstract class AbstractTrainingExampleExtractor implements ITrainingExampleExtractor{
	private static final long serialVersionUID = 1L;


	/*
	 * When generating pairwise examples, we have the problem that sometimes the
	 * closest preceding antecedent of the anaphor is not in the predspans list
	 * What to do then?
	 * 1) Always commit the instances;
	 * 2) Commit only when there is a positive somewhere;
	 * 3) Commit only when the closest preceding antecedent is in the list
	 */
	public static enum CommitStrategy { Always, OnlyWhenPositiveExists, OnlyWhenImmediatePrecedingExists }
	
	protected final IMarkableExtractor markableExtractor;
	protected final CommitStrategy commitStrategy;
	
	protected AbstractTrainingExampleExtractor(IMarkableExtractor markableExtractor,CommitStrategy commitStrategy){
		this.markableExtractor=markableExtractor;
		this.commitStrategy=commitStrategy;
	}

//	protected Map<Integer,Chain> getGoldChainMap(Document d){
//		GoldStandardChainExtractor goldExtractor=new GoldStandardChainExtractor();
//		Map<Integer,Chain> chainMap=goldExtractor.getGoldChains(d);
//		return chainMap;
//	}
	
	protected Chain[] getGoldChains(Document d){
		GoldStandardChainExtractor goldExtractor=new GoldStandardChainExtractor();
		Chain[] gold=goldExtractor.getGoldChains(d);
		return gold;
	}
	
//	private Map<Span,Integer> getSpan2IntMap(Iterable<Chain> cit){
	private Map<Span,Integer> getSpan2IntMap(Chain[] cit){
		Map<Span,Integer> span2int=new HashMap<Span,Integer>();
		for(Chain c:cit){
			if(c!=null)
				for(Span s:c.spans)
					span2int.put(s, c.chainId);
		}
		return span2int;
	}
	
//	private Map<Span,Integer> getSpan2IntMap(Map<Integer,Chain> chainMap){
//		Map<Span,Integer> span2int=new HashMap<Span,Integer>();
//		for(Entry<Integer,Chain> e:chainMap.entrySet()){
//			for(Span s:e.getValue().spans)
//				span2int.put(s, e.getKey());
//		}
//		return span2int;
//	}
	
	private static void pruneVerbsFromChains(Chain[] goldChains) {
		for(int i=0;i<goldChains.length;++i){
			Chain c=goldChains[i];
			Iterator<Span> sIt=c.spans.iterator();
			while(sIt.hasNext()){
				Span s=sIt.next();
				if(s.s.tags[s.hd].startsWith("V"))
					sIt.remove();
			}
			if(c.spans.size()<2)
				goldChains[i]=null;
		}
	}
	
//	@Deprecated
//	private static Map<Integer,Chain> pruneVerbsFromChains(Map<Integer,Chain> chainMap){
//		Iterator<Entry<Integer,Chain>> it=chainMap.entrySet().iterator();
//		while(it.hasNext()){
//			Entry<Integer,Chain> e=it.next();
//			Iterator<Span> sIt=e.getValue().spans.iterator();
//			while(sIt.hasNext()){
//				Span s=sIt.next();
//				if(s.s.tags[s.hd].startsWith("V"))
//					sIt.remove();
//			}
//			if(e.getValue().spans.size()<2)
//				it.remove();
//		}
//		return chainMap;
//	}
	
//	@Override
//	public List<PairInstance> getInstances(Document d) {
//		Set<Span> predSpans=markableExtractor.extractMarkables(d);
//		Map<Integer,Chain> goldMap=getGoldChainMap(d);
//		return getInstances(goldMap,predSpans);
//	}
	
//	private List<PairInstance> getInstances(Chain[] goldChains,Set<Span> predSpans){
	public List<PairInstance> getInstances(Document d){
		Chain[] goldChains=getGoldChains(d);
		Set<Span> predSpans=markableExtractor.extractMarkables(d);
		SpanListStruct sls=SpanListStruct.fromCollection(predSpans);
		pruneVerbsFromChains(goldChains);
		Map<Span,Integer> goldSpan2Int=getSpan2IntMap(goldChains);
		List<PairInstance> r=new ArrayList<PairInstance>();
		for(Chain c:goldChains)
			if(c!=null)
				extractExamplesByChain(goldSpan2Int,c,sls,r);
		return r;
	}

//	@Deprecated
//	private List<PairInstance> getInstances(Map<Integer, Chain> goldChains, Set<Span> predSpans){
//		SpanListStruct sls=SpanListStruct.fromCollection(predSpans);
//		pruneVerbsFromChains(goldChains);
//		Map<Span,Integer> goldSpan2Int=getSpan2IntMap(goldChains.values());
//		List<PairInstance> r=new ArrayList<PairInstance>();
//		for(Entry<Integer,Chain> e:goldChains.entrySet()){
//			extractExamplesByChain(goldSpan2Int,e.getValue(),sls,r);
//		}
//		return r;
//	}
	

	protected boolean skipAnaphor(Chain ch, SpanListStruct sls,	int anaChainIndex) {
		if(commitStrategy==CommitStrategy.OnlyWhenImmediatePrecedingExists){
			Span closestPreceedingGoldAnt=ch.spans.get(anaChainIndex-1);
			if(!sls.contains(closestPreceedingGoldAnt))
				return true;
			else
				return false;
		} else if(commitStrategy==CommitStrategy.OnlyWhenPositiveExists){
			boolean con=true;
			for(int r=anaChainIndex-1;r>=0;--r){
				Span s=ch.spans.get(r);
				if(sls.contains(s)){
					con=false;
					break;
				}
			}
			return con;
		} else {
			return false;
		}
	}

	protected static boolean sameChain(Span ant, Span ana, Map<Span, Integer> span2int) {
		Integer antI=span2int.get(ant);
		Integer anaI=span2int.get(ana);
		return antI!=null && anaI!=null && antI.equals(anaI);
	}

	
//	abstract void extractExamplesByChain(Map<Span,Integer> goldSpan2Int,Chain ch,List<Span> predSpans,List<PairInstance> sink);
	abstract void extractExamplesByChain(Map<Span,Integer> goldSpan2Int,Chain ch,SpanListStruct sls,List<PairInstance> sink);
}
